/*

JuggleFire
by Brent P. Newhall <brent@stun-software.com>

This is a free demonstration of a scrolling starfield.  This demo uses a Thread
to send Messages to the window every microsecond.  Each message moves the stars
by one step.  If a star moves off the left edge of the window, it is re-spawned
at the right edge of the window.

The stars are implemented as individual classes derived from Point.  Each star
has a depth from 0-2.  0 is nearer and brighter.

The starfield is double-buffered using Bitmaps.

 */

#include <gui/window.h>
#include <gui/view.h>
#include <gui/rect.h>
#include <gui/imagebutton.h>
#include <util/application.h>
#include <vector>
#include <atheos/random.h>
#include <util/thread.h>
#include <util/message.h>
#include <storage/file.h>
#include <gui/image.h>
#include <gui/imageview.h>
#include <gui/scrollbar.h>
#include <media/soundplayer.h>

using namespace os;




const unsigned int NUM_STARS = 250;
const unsigned int MSG_MOVE_STARS = 1001;

const unsigned int WEAPON_SMALL = 1;
const unsigned int WEAPON_BIG   = 2;
const unsigned int WEAPON_WAVE  = 3;

Window *window;




// ---------- Bullet



class Bullet
	{
	public:
		unsigned int type;
		Color32_s color;
		float velocity;
		Point location;
	};



// ---------- Menu Bar

class MenuBarView : public View
	{
	public:
		MenuBarView( const Rect& r );
	private:
		ImageButton *enemy1Button;
		ImageButton *enemy2Button;
		ImageButton *enemy3Button;
	};




MenuBarView :: MenuBarView( const Rect& r )
	: View( r, "MenuBarView", CF_FOLLOW_LEFT | CF_FOLLOW_TOP, WID_WILL_DRAW | WID_CLEAR_BACKGROUND )
	{
	Rect size( 0, 0, 63, 63 );
	BitmapImage *img;
	File *f;
	f = new File( "^/graphics/player_red.png" );
	img = new BitmapImage( f );
	enemy1Button = new ImageButton( size, "Button1", "Enemy 1", NULL, img, ImageButton::IB_TEXT_BOTTOM, true );
	AddChild( enemy1Button );
	f = new File( "^/graphics/player_yellow.png" );
	size.left += 70;
	size.right += 70;
	img = new BitmapImage( f );
	enemy2Button = new ImageButton( size, "Button2", "Enemy 2", NULL, img, ImageButton::IB_TEXT_BOTTOM, true );
	AddChild( enemy2Button );
	f = new File( "^/graphics/player_red.png" );
	size.left += 70;
	size.right += 70;
	img = new BitmapImage( f );
	enemy3Button = new ImageButton( size, "Button3", "Enemy 3", NULL, img, ImageButton::IB_TEXT_BOTTOM,  true );
	AddChild( enemy3Button );
	}



// ---------- Main View



class EditorView : public View
	{
	public:
		EditorView( const Rect& r ); /* Constructor */
		void Paint( const Rect& updateRect );
		void DrawStars( void );
		void KeyHit( int32 nRawKey );
		void fireWeapon( void );
		void updateShipType( void );
	private:
		void LoadBitmaps( void );
		Color32_s color[3];
		View* drawingView;
		Bitmap* offscreenBitmap;
		Bitmap* onscreenBitmap;
		Point playerLocation;
		unsigned int playerWeaponType;
		unsigned int playerShipType;
		BitmapImage *playerRedImage;
		BitmapImage *playerYellowImage;
		BitmapImage *playerImage;
		std::vector<Bullet> bullets;
		unsigned int chargeDelay;
		MediaSoundPlayer *sound;
	};




EditorView :: EditorView( const Rect& r )
	: View( r, "EditorView", CF_FOLLOW_LEFT | CF_FOLLOW_TOP, WID_WILL_DRAW | WID_CLEAR_BACKGROUND )
	{
	unsigned int i;
	/* Misc */
	chargeDelay = 0;
	LoadBitmaps();
	/* Initialize colors */
	color[0].red = 50;
	color[0].green = 50;
	color[0].blue = 50;
	color[1].red = 150;
	color[1].green = 150;
	color[1].blue = 150;
	color[2].red = 250;
	color[2].green = 250;
	color[2].blue = 250;
	/* Initialize player information */
	playerLocation.x = 15;
	playerLocation.y = 40;
	playerWeaponType = 1;
	playerShipType = 1;
	updateShipType();
	/* Initialize double-buffering controls */
	drawingView = new View( r, "DrawView" );
	offscreenBitmap = new Bitmap( (int)r.Width() + 1, (int)r.Height() + 1,
	                             CS_RGB32,
	                             Bitmap::ACCEPT_VIEWS | Bitmap::SHARE_FRAMEBUFFER );
	offscreenBitmap->AddChild( drawingView );
	onscreenBitmap = new Bitmap( (int)r.Width() + 1, (int)r.Height() + 1, CS_RGB32 );
	/* Play initial sound */
	/* sound = new MediaSoundPlayer();
	sound->SetFile( "^/sounds/sucks.wav" );
	sound->Play(); */
	}




void EditorView :: LoadBitmaps( void )
	{
	File *f;
	/* Load red player sprite */
	f = new File( "^/graphics/player_red1.png" );
	playerRedImage = new BitmapImage( f );
	/* Load yellow player sprite */
	f = new File( "^/graphics/player_yellow1.png" );
	playerYellowImage = new BitmapImage( f );
	}




void EditorView :: Paint( const Rect& updateRect )
	{
	DrawBitmap( onscreenBitmap, updateRect, updateRect );
	}




void EditorView :: DrawStars( void )
	{
	unsigned int i;
	Rect bounds = GetBounds();
	drawingView->SetFgColor( 0, 0, 0 );
	drawingView->FillRect( bounds );
	/* Draw player */
	drawingView->SetDrawingMode( DM_COPY );
	playerImage->Draw( playerLocation, drawingView );
	/* Draw bullets */
	for( i = 0; i < bullets.size(); i++ )
		{
		bullets[i].location.x += bullets[i].velocity;
		drawingView->SetFgColor( bullets[i].color );
		Rect r( bullets[i].location.x, bullets[i].location.y, bullets[i].location.x+2, bullets[i].location.y+2 );
		if( bullets[i].type == WEAPON_BIG )
			{
			r.bottom += 3;
			r.right += 3;
			}
		else if( bullets[i].type == WEAPON_WAVE )
			{
			r.bottom += 20;
			}
		drawingView->FillRect( r );
		if( r.left <= 0  ||  r.right >= bounds.Width() )
			{
			std::vector<Bullet>::iterator it;
			it = bullets.begin() + i;
			bullets.erase( it );
			}
		}
	/* Update view */
	drawingView->Sync();
	memcpy( onscreenBitmap->LockRaster(),
	        offscreenBitmap->LockRaster(),
	        onscreenBitmap->GetBytesPerRow() *
	          ((int)onscreenBitmap->GetBounds().Height() + 1) );
	Invalidate();
	Sync();
	/* Update charge delay */
	if( chargeDelay > 0 )
		chargeDelay--;
	}




void EditorView :: KeyHit( int32 nKeyCode )
	{
//printf( "%x\n", nKeyCode );
	const float playerMovementIncrement = 5.0;
	if( nKeyCode == 0x57 ) /* Up arrow */
		{
		playerLocation.y -= playerMovementIncrement;
		if( playerLocation.y < 10.0 )  playerLocation.y = 10.0;
		}
	else if( nKeyCode == 0x62	 ) /* Down arrow */
		{
		playerLocation.y += playerMovementIncrement;
		if( playerLocation.y > GetBounds().Height() - 20.0 )  playerLocation.y = GetBounds().Height() - 10.0;
		}
	else if( nKeyCode == 0x61 ) /* Left arrow */
		{
		playerLocation.x -= playerMovementIncrement;
		if( playerLocation.x < 10.0 )  playerLocation.x = 10.0;
		}
	else if( nKeyCode == 0x63 ) /* Right arrow */
		{
		playerLocation.x += playerMovementIncrement;
		if( playerLocation.x > GetBounds().Width() - 20.0 )  playerLocation.x = GetBounds().Width() - 10.0;
		}
	else if( nKeyCode == 0x5e ) /* Space */
		{
		fireWeapon();
		}
	else if( nKeyCode == 0x3c ) /* A */
		{
		playerWeaponType = WEAPON_SMALL;
		}
	else if( nKeyCode == 0x3d ) /* S */
		{
		playerWeaponType = WEAPON_BIG;
		}
	else if( nKeyCode == 0x3e ) /* D */
		{
		playerWeaponType = WEAPON_WAVE;
		}
	else if( nKeyCode == 0x3f ) /* F */
		{
		if( playerShipType == 1 )
			playerShipType = 2;
		else
			playerShipType = 1;
		updateShipType();
		}
	}




void EditorView :: fireWeapon( void )
	{
	if( chargeDelay > 0 )
		return;
	Bullet b;
	Color32_s c( 250, 50, 50 ); /* Default bullet to red */
	b.location.x = playerLocation.x + 10;
	b.location.y = playerLocation.y + 3;
	b.type = playerWeaponType;
	if( b.type == WEAPON_WAVE )
		b.location.y -= 10;
	if( playerShipType == 2 ) /* Turn bullet yellow if ship is in yellow mode */
		c.green = 250;
	b.color = c;
	b.velocity = 3.0;
	bullets.push_back( b );
	if( playerWeaponType == WEAPON_SMALL )
		chargeDelay = 10;
	else if( playerWeaponType == WEAPON_BIG )
		chargeDelay = 30;
	else
		chargeDelay = 50;
	}




void EditorView :: updateShipType( void )
	{
	/* Toggle between red and yellow modes */
	if( playerShipType == 1 )
		{
		playerImage = playerRedImage;
		}
	else
		{
		playerImage = playerYellowImage;
		}
	}




// ---------- Stars Window



class EditorWindow : public Window
	{
	public:
		EditorWindow( const Rect& r ); /* Constructor */
		void HandleMessage( Message* pcMessage );
		void DispatchMessage( Message* pcMessage, Handler *pcHandler );
		bool OkToQuit();
	private:
		EditorView *view;
		MenuBarView *menuBar;
		ScrollBar *scrollBar;
	};




EditorWindow :: EditorWindow( const Rect& r )
	: Window( r, "EditorWindow", "Juggle Fire", WND_NOT_RESIZABLE | WND_NO_ZOOM_BUT | WND_NO_DEPTH_BUT, CURRENT_DESKTOP )
	{
	window = this;
	menuBar = new MenuBarView( Rect( 0, 0, GetBounds().Width(), 70 ) );
	AddChild( menuBar );
	view = new EditorView( Rect( 0, 71, 10000, GetBounds().Height() - 30 ) );
	AddChild( view );
	scrollBar = new ScrollBar( Rect( 0, GetBounds().Height() - 30, GetBounds().Width(), GetBounds().Height() ), "WinScrollBar", NULL, 0, 1000, HORIZONTAL, CF_FOLLOW_ALL );
	scrollBar->SetSteps( 5, 50 );
	scrollBar->SetScrollTarget( view );
	AddChild( scrollBar );
	//SetFocusChild( view );
	}




void EditorWindow :: HandleMessage( Message *pcMessage )
	{
	switch( pcMessage->GetCode() )
		{
		case MSG_MOVE_STARS:
			view->DrawStars();
			break;
		default:
			Window :: HandleMessage( pcMessage );
		}
	}




void EditorWindow :: DispatchMessage( Message* pcMessage, Handler* pcHandler )
	{
/*	String cRawString; */
	int32 /* nQual, */ nRawKey;
	if( pcMessage->GetCode() == M_KEY_DOWN )
		{
/*		pcMessage->FindString( "_raw_string", &cRawString );
		pcMessage->FindInt32( "_qualifiers", &nQual ); */
		pcMessage->FindInt32( "_raw_key", &nRawKey );
		view->KeyHit( nRawKey );
		}
	Window::DispatchMessage( pcMessage, pcHandler );
	}




bool EditorWindow :: OkToQuit( void )
	{
	Application::GetInstance()->PostMessage( M_QUIT );
	return( true );
	}




// ---------- Stars Application



class EditorApplication : public Application
	{
	public:
		EditorApplication( void );
	private:
		EditorWindow *win;
	};




EditorApplication :: EditorApplication( void )
		: Application( "application/x-vind.STUN-JuggleFire" )
	{
	win = new EditorWindow( Rect(20,40,700,380) );
	win->Show();
	win->MakeFocus();
	}




// ---------- main()




int main( void )
	{
	EditorApplication *thisApp;
	thisApp = new EditorApplication();
	thisApp->Run();
	return( 0 );
	}

