/*

Firestorm
by Brent P. Newhall <brent@stun-software.com>

 */

#include <gui/window.h>
#include <gui/view.h>
#include <gui/rect.h>
#include <util/application.h>
#include <vector>
#include <algorithm>
#include <atheos/random.h>
#include <util/thread.h>
#include <util/message.h>
#include <storage/file.h>
#include <storage/filereference.h>
#include <gui/image.h>
#include <gui/imageview.h>
#include <media/soundplayer.h>
#include <fstream>

using namespace os;




const unsigned int NUM_STARS = 250;
const unsigned int MSG_MOVE_STARS        = 1001;
const unsigned int MSG_FIRE_ENEMY_BULLET = 1002;

const unsigned int WEAPON_SMALL = 1;
const unsigned int WEAPON_BIG   = 2;
const unsigned int WEAPON_WAVE  = 3;

Window *window;




// ---------- Utilities




float calc_float_number( const float number_part1, const float number,
                         bool is_negative )
	{
	float final_number;
	if( number_part1 == 0 )
		final_number = (float)number;
	else
		final_number = (float)number_part1 + ((float)number * 0.1);
	if( is_negative )
		final_number = 0 - final_number;
	return final_number;
	}




std::vector<float> * parse_line( const char * line )
	{
	std::vector<float> *v;
	v = new std::vector<float>;
	unsigned int i, number = 0, number_part1 = 0;
	char c;
	bool in_number = false;
	bool is_negative = false;
	for( i = 0; i < strlen( line ); i++ )
		{
		c = line[i];
		if( c >= 48  &&  c <= 57 ) // Is a digit
			{
			if( in_number )
				number = (number * 10) + (c - 48);
			else
				{
				in_number = true;
				number = (c - 48);
				}
			}
		else if( c == '.' )
			{
			if( in_number )
				{
				number_part1 = number;
				number = 0;
				}
			}
		else if( c == '-' )
			{
			is_negative = true;
			}
		else
			{
			if( in_number )
				{
				in_number = false;
				v->push_back( calc_float_number( number_part1, number,
				              is_negative ) );
				number = 0;
				number_part1 = 0;
				}
			is_negative = false;
			}
		}
	if( in_number )
		v->push_back( calc_float_number( number_part1, number, is_negative ) );
	return v;
	}




// ---------- Move Stars Thread




class MoveStarsThread : public Thread
	{
	public:
		MoveStarsThread() : Thread( "MoveStarsThread", LOW_PRIORITY ) { }
		int32 Run()
			{
			while( 1 == 1 )
				{
				Delay( 1000 ); /* Pause for one microsecond. */
				window->PostMessage( new Message( MSG_MOVE_STARS ), window );
				}
			return( 0 );
			}
	};




class Star : public Point
	{
	public:
		Star();
		void PlaceRandomly( void );
		unsigned int depth;
	};




Star :: Star()
	{
	x = 0;
	y = 0;
	depth = 0;
	}




void Star :: PlaceRandomly( void )
	{
	x = rand() % (int)(window->GetBounds().Width());
	y = rand() % (int)(window->GetBounds().Height());
	depth = rand() % 3;
	}




// ---------- Bullet




class Bullet
	{
	public:
		unsigned int type;
		Color32_s color;
		Point velocity;
		Point location;
		unsigned int int_color;
	};




// ---------- Enemy




class EnemyMovementEvent
	{
	public:
		EnemyMovementEvent()
			{
			time = 0;
			velocity.x = -3.0;
			velocity.y = 0.0;
			}
		unsigned int time;
		Point velocity;
	};




class Enemy
	{
	public:
		Enemy();
		void SetPath( std::vector<float> *pathNumbers );
		void update( void );
		unsigned int type;
		BitmapImage *image;
		Rect location;
		Point velocity;
		float direction;
		unsigned int color;
		unsigned int pathTicks;
		std::vector<EnemyMovementEvent> movements;
	private:
		unsigned int heartbeat;
		unsigned int nextMovement;
	};




Enemy :: Enemy()
	{
	type = 0;
	location.left = 0;
	location.top = 0;
	location.right = 63;
	location.bottom = 63;
	velocity.x = 0.0;
	velocity.y = 0.0;
	direction = 0.0;
	color = 1;
	pathTicks = 30;
	heartbeat = 0;
	nextMovement = 0;
	}




void Enemy :: SetPath( std::vector<float> *pathNumbers )
	{
	Enemy();
	unsigned int i;
	float direction;
	float velocity;
	EnemyMovementEvent event;
	movements.clear();
	/* We grab the last element in each tuple and the two before it;
	   otherwise, we might get near the end of an uneven vector and try to
	   look beyond the last item. */
	for( i = 2; i < pathNumbers->size(); i = i + 3 )
		{
		event.time = (int)((*pathNumbers)[i - 2]);
		event.velocity.x  = (*pathNumbers)[i - 1];
		event.velocity.y  = (*pathNumbers)[i];
		movements.push_back( event );
		}
	}




void Enemy :: update( void )
	{
	heartbeat++;
	unsigned int pathTick = heartbeat % pathTicks;
	if( movements[nextMovement].time == pathTick )
		{
		/* Activate the next movement */
		if( movements[nextMovement].velocity.x == 0.0  &&
		    movements[nextMovement].velocity.y == 0.0 )
			{
			Message msg( MSG_FIRE_ENEMY_BULLET );
			msg.AddPoint( "location", Point( location.left, location.top - 32 ) );
			msg.AddInt32( "type", (int32)color );
			window->PostMessage( &msg );
			}
		velocity = movements[nextMovement].velocity;
		if( nextMovement == movements.size() - 1 )
			nextMovement = 0;
		else
			nextMovement++;
		}
	location.left   += velocity.x;
	location.right  += velocity.x;
	location.top    += velocity.y;
	location.bottom += velocity.y;
	}




// ---------- Event Queue




class LevelEvent
	{
	public:
		LevelEvent();
		int operator< ( const LevelEvent &x ) const
		{  return( this->time < x.time );  }
		unsigned int time;
		float location;
		unsigned int enemyType;
		String path;
		unsigned int color;
	};




LevelEvent :: LevelEvent()
	{
	time = 0;
	location = 40.0;
	enemyType = 1;
	color = 1;
	}




class LevelEventQueue
	{
	public:
		LevelEventQueue();
		std::vector<LevelEvent *> events;
	};




LevelEventQueue :: LevelEventQueue()
	{
	std::vector<float> *lineOfNumbers;
	FileReference ref( Directory( "^/levels" ), "level1" );
	String path;
	ref.GetPath( &path );
	std::fstream levelFile( path.c_str(), std::ios::in );
	String line;
	char cline[4096];
	bool inPaths = true;
	while( ! levelFile.eof() )
		{
		levelFile.getline( cline, 4096 );
		line = (String)cline;
		if( line.substr( 0, 7 ) == "[paths]" )
			inPaths = true;
		else if( line.substr( 0, 7 ) == "[ships]" )
			inPaths = false;
		else if( inPaths )
			{
			//lineOfNumbers = parse_line( cline );
			}
		else if( ! inPaths ) // In Ships section
			{
			lineOfNumbers = parse_line( cline );
			}
		}
	levelFile.close();
	LevelEvent *e = new LevelEvent();
	e->time = 100;
	e->location = 70.0;
	e->enemyType = 1;
	e->color = 2;
	e->path = "1  50  1,-5,0";
	events.push_back( e );
	e = new LevelEvent();
	e->time = 150;
	e->location = 120.0;
	e->enemyType = 2;
	e->color = 2;
	e->path = "1  50  1,-5,0";
	events.push_back( e );
	e = new LevelEvent();
	e->time = 250;
	e->location = 160.0;
	e->enemyType = 3;
	e->color = 2;
	e->path = "1  50  1,-5,0";
	events.push_back( e );
	sort( events.begin(), events.end() );
	}




// ---------- Stars View



class StarsView : public View
	{
	public:
		StarsView( const Rect& r ); /* Constructor */
		~StarsView();
		void Paint( const Rect& updateRect );
		void DrawStars( void );
		void UpdateAndDrawEnemies( void );
		void KeyUp( const char *pzString, const char *pzRawString,
		            uint32 nQualifiers );
		void KeyDown( const char *pzString, const char *pzRawString,
		              uint32 nQualifiers );
		void KeyHit( int32 nRawKey );
		void fireWeapon( void );
		void fireEnemyBullet( Point location, unsigned int type );
		void updateShipType( void );
	private:
		unsigned int levelTimeCount;
		void LoadBitmaps( void );
		Star stars[NUM_STARS];
		Color32_s color[3];
		View* drawingView;
		Bitmap* offscreenBitmap;
		Bitmap* onscreenBitmap;
		Point playerLocation;
		unsigned int playerWeaponType;
		unsigned int playerShipType;
		BitmapImage *playerRedImage;
		BitmapImage *playerYellowImage;
		BitmapImage *playerImage;
		std::vector<Bullet> bullets;
		std::vector<Enemy> enemies;
		unsigned int chargeDelay;
		LevelEventQueue *eventQueue;
		MediaManager *pcManager;
		MediaSoundPlayer *sound;
		bool holdingFireButton1;
		bool holdingFireButton2;
		bool holdingFireButton3;
		bool holdingUpArrow;
		bool holdingDownArrow;
		bool holdingLeftArrow;
		bool holdingRightArrow;
	};




StarsView :: StarsView( const Rect& r )
	: View( r, "StarsView", CF_FOLLOW_LEFT | CF_FOLLOW_TOP,
	        WID_WILL_DRAW | WID_CLEAR_BACKGROUND )
	{
	unsigned int i;
	/* Misc */
	levelTimeCount = 0;
	chargeDelay = 0;
	LoadBitmaps();
	/* Initialize colors */
	color[0].red = 50;
	color[0].green = 50;
	color[0].blue = 50;
	color[1].red = 150;
	color[1].green = 150;
	color[1].blue = 150;
	color[2].red = 250;
	color[2].green = 250;
	color[2].blue = 250;
	/* Initialize stars */
	for( i = 0; i < NUM_STARS; i++ )
		{
		stars[i].PlaceRandomly();
		}
	/* Initialize keyboard key flags */
	holdingFireButton1 = false;
	holdingFireButton2 = false;
	holdingFireButton3 = false;
	holdingUpArrow     = false;
	holdingDownArrow   = false;
	holdingLeftArrow   = false;
	holdingRightArrow  = false;
	/* Initialize player information */
	playerLocation.x = 15;
	playerLocation.y = 40;
	playerWeaponType = 1;
	playerShipType = 1;
	updateShipType();
	/* Initialize double-buffering controls */
	drawingView = new View( r, "DrawView" );
	offscreenBitmap = new Bitmap( (int)r.Width() + 1, (int)r.Height() + 1,
	                             CS_RGB32,
	                             Bitmap::ACCEPT_VIEWS |
	                             Bitmap::SHARE_FRAMEBUFFER );
	offscreenBitmap->AddChild( drawingView );
	onscreenBitmap = new Bitmap( (int)r.Width() + 1, (int)r.Height() + 1, CS_RGB32 );
	eventQueue = new LevelEventQueue();
	/* Play initial sound */
	pcManager = MediaManager::Get();
	sound = new MediaSoundPlayer();
	FileReference ref( Directory( "^/sounds" ), "Bullet.ogg" );
	String path;
	ref.GetPath( &path );
	sound->SetFile( path );
	sound->Play();
	}




StarsView :: ~StarsView()
	{
	delete playerRedImage;
	delete playerYellowImage;
	delete offscreenBitmap;
	delete onscreenBitmap;
	pcManager->Put();
	}




void StarsView :: LoadBitmaps( void )
	{
	File *f;
	/* Load red player sprite */
	f = new File( "^/graphics/player_red1.png" );
	playerRedImage = new BitmapImage( f );
	/* Load yellow player sprite */
	f = new File( "^/graphics/player_yellow1.png" );
	playerYellowImage = new BitmapImage( f );
	}




void StarsView :: Paint( const Rect& updateRect )
	{
	DrawBitmap( onscreenBitmap, updateRect, updateRect );
	}




void StarsView :: DrawStars( void )
	{
	unsigned int i;
	levelTimeCount++; /* It's been another tick in the level's life */
	/* Draw black background */
	Rect bounds = GetBounds();
	drawingView->SetFgColor( 0, 0, 0 );
	drawingView->FillRect( bounds );
	/* Draw stars */
	for( i = 0; i < NUM_STARS; i++ )
		{
		stars[i].x -= stars[i].depth + 1;
		if( stars[i].x < 0 )
			{
			stars[i].PlaceRandomly();
			stars[i].x = bounds.Width();
			}
		drawingView->SetFgColor( color[stars[i].depth] );
		drawingView->DrawLine( stars[i], stars[i] );
		}
	/* Draw enemies */
	UpdateAndDrawEnemies();
	/* Draw player */
	KeyHit( 0 );
	drawingView->SetDrawingMode( DM_BLEND );
	playerImage->Draw( playerLocation, drawingView );
	drawingView->SetDrawingMode( DM_COPY );
	/* Draw bullets */
	for( i = 0; i < bullets.size(); i++ )
		{
		bullets[i].location.x += bullets[i].velocity.x;
		bullets[i].location.y += bullets[i].velocity.y;
		drawingView->SetFgColor( bullets[i].color );
		Rect r( bullets[i].location.x, bullets[i].location.y,
		        bullets[i].location.x + 2, bullets[i].location.y + 2 );
		if( bullets[i].type == WEAPON_WAVE )
			{
			r.top -= 20;
			r.bottom += 35;
			}
		drawingView->FillRect( r );
		if( r.left <= 0  ||  r.right >= bounds.Width() )
			{
			std::vector<Bullet>::iterator it;
			it = bullets.begin() + i;
			bullets.erase( it );
			}
		}
	/* Update view */
	drawingView->Sync();
	memcpy( onscreenBitmap->LockRaster(),
	        offscreenBitmap->LockRaster(),
	        onscreenBitmap->GetBytesPerRow() *
	          ((int)onscreenBitmap->GetBounds().Height() + 1) );
	Invalidate();
	Sync();
	/* Update charge delay */
	if( chargeDelay > 0 )
		chargeDelay--;
	}




void StarsView :: UpdateAndDrawEnemies( void )
	{
	unsigned int i, j;
	/* Move existing enemies */
	for( i = 0; i < enemies.size(); i++ )
		{
		enemies[i].update();
		/* Remove this enemy if it goes off the screen */
		if( enemies[i].location.right <= 1 )
			{
			enemies.erase( enemies.begin() + i );
			i--;
			}
		}
	/* Check for collisions between bullets and enemies */
	for( i = 0; i < bullets.size(); i++ )
		{
		for( j = 0; i < bullets.size()  &&  j < enemies.size(); j++ )
			{
			if( enemies[j].location.DoIntersect( bullets[i].location )  &&
			    enemies[j].color == bullets[i].int_color )
				{
				delete enemies[j].image;
				enemies.erase( enemies.begin() + j );
				j--;
				bullets.erase( bullets.begin() + i );
				i--;
				}
			}
		}
	/* Check for new enemies in queue */
	while( eventQueue->events.size() > 0  &&
	       eventQueue->events[0]->time == levelTimeCount )
		{
		Enemy enemy;
		enemy.type = eventQueue->events[0]->enemyType;
		String filename = "^/graphics/enemy";
		char enemyTypeString[2];
		sprintf( enemyTypeString, "%d", enemy.type );
		filename += (String)(enemyTypeString) + "_";
		if( eventQueue->events[0]->color == 1 )
			filename += "red";
		else
			filename += "yellow";
		filename += ".png";
		enemy.image = new BitmapImage( new File( filename ) );
		enemy.location.left = GetBounds().Width();
		enemy.location.top = eventQueue->events[0]->location;
		enemy.location.right = enemy.location.left + 63;
		enemy.location.bottom = enemy.location.top + 63;
		enemy.direction = -0.5;
		enemy.color = eventQueue->events[0]->color;
		std::vector<float> pathNumbers;
		std::vector<float> *v = parse_line(
		   eventQueue->events[0]->path.c_str() );
		enemy.pathTicks = (unsigned int)((*v)[1]);
		for( i = 2; i < (*v).size(); i++ )
			{
			pathNumbers.push_back( (*v)[i] );
			}
		enemy.SetPath( &pathNumbers );
		enemies.push_back( enemy );
		eventQueue->events.erase( eventQueue->events.begin() );
		}
	/* Draw all existing enemies */
	drawingView->SetDrawingMode( DM_BLEND );
	for( i = 0; i < enemies.size(); i++ )
		{
		enemies[i].image->Draw( Point( enemies[i].location.left,
		                               enemies[i].location.top ), drawingView );
		}
	drawingView->SetDrawingMode( DM_COPY );
	}




void StarsView :: KeyUp( const char *pzString, const char *pzRawString,
                         uint32 nQualifiers )
	{
	int32 nRawKey;
	GetWindow()->GetCurrentMessage()->FindInt32( "_raw_key", &nRawKey );
	if( nRawKey == 0xD7 ) /* Up arrow */
		holdingUpArrow = false;
	else if( nRawKey == 0xE2 ) /* Down arrow */
		holdingDownArrow = false;
	else if( nRawKey == 0xE1 ) /* Left arrow */
		holdingLeftArrow = false;
	else if( nRawKey == 0xE3 ) /* Right arrow */
		holdingRightArrow = false;
	else if( pzRawString[0] == 'f' )
		holdingFireButton1 = false;
	else if( pzRawString[0] == 'd' )
		holdingFireButton2 = false;
	else if( pzRawString[0] == 's' )
		holdingFireButton3 = false;
	}




void StarsView :: KeyDown( const char *pzString, const char *pzRawString,
                           uint32 nQualifiers )
	{
	int32 nRawKey;
	GetWindow()->GetCurrentMessage()->FindInt32( "_raw_key", &nRawKey );
	if( nRawKey == 0x57 ) /* Up arrow */
		holdingUpArrow = true;
	else if( nRawKey == 0x62 ) /* Down arrow */
		holdingDownArrow = true;
	else if( nRawKey == 0x61 ) /* Left arrow */
		holdingLeftArrow = true;
	else if( nRawKey == 0x63 ) /* Right arrow */
		holdingRightArrow = true;
	else if( pzRawString[0] == 'f' )
		holdingFireButton1 = true;
	else if( pzRawString[0] == 'd' )
		holdingFireButton2 = true;
	else if( pzRawString[0] == 's' )
		holdingFireButton3 = true;
	else if( pzRawString[0] == ' ' )
		{
		if( playerShipType == 1 )
			playerShipType = 2;
		else
			playerShipType = 1;
		updateShipType();
		}
	}




void StarsView :: KeyHit( int32 nKeyCode )
	{
	const float playerMovementIncrement = 5.0;
	if( holdingUpArrow ) /* Up arrow */
		{
		playerLocation.y -= playerMovementIncrement;
		if( playerLocation.y < 10.0 )  playerLocation.y = 10.0;
		}
	if( holdingDownArrow ) /* Down arrow */
		{
		playerLocation.y += playerMovementIncrement;
		if( playerLocation.y > GetBounds().Height() - 60.0 )
			playerLocation.y = GetBounds().Height() - 50.0;
		}
	if( holdingLeftArrow ) /* Left arrow */
		{
		playerLocation.x -= playerMovementIncrement;
		if( playerLocation.x < 10.0 )  playerLocation.x = 10.0;
		}
	if( holdingRightArrow ) /* Right arrow */
		{
		playerLocation.x += playerMovementIncrement;
		if( playerLocation.x > GetBounds().Width() - 60.0 )
			playerLocation.x = GetBounds().Width() - 50.0;
		}
	if( holdingFireButton3 )
		{
		playerWeaponType = WEAPON_WAVE;
		fireWeapon();
		}
	if( holdingFireButton2 )
		{
		playerWeaponType = WEAPON_BIG;
		fireWeapon();
		}
	else if( holdingFireButton1 )
		{
		playerWeaponType = WEAPON_SMALL;
		fireWeapon();
		}
	}




void StarsView :: fireWeapon( void )
	{
	if( chargeDelay > 0 )
		return;
	Bullet b;
	Color32_s c( 250, 50, 50 ); /* Default bullet to red */
	b.int_color = 1;
	b.location.x = playerLocation.x + 55;
	b.location.y = playerLocation.y + 32;
	b.type = playerWeaponType;
	if( b.type == WEAPON_WAVE )
		b.location.y -= 10;
	if( playerShipType == 2 ) /* Turn bullet yellow if ship is in yellow mode */
		{
		c.green = 250;
		b.int_color = 2;
		}
	b.color = c;
	b.velocity.x = 5.0;
	b.velocity.y = 0.0;
	if( playerWeaponType == WEAPON_SMALL )
		chargeDelay = 10;
	else if( playerWeaponType == WEAPON_BIG )
		chargeDelay = 30;
	else
		chargeDelay = 50;
	bullets.push_back( b );
	if( playerWeaponType == WEAPON_BIG )
		{
		/* Add two more bullets */
		Bullet b1, b2;
		b1 = b;
		b2 = b;
		b1.velocity.x = 4;
		b1.velocity.y = -2.0;
		b2.velocity.x = 4;
		b2.velocity.y = 2.0;
		bullets.push_back( b1 );
		bullets.push_back( b2 );
		}
	sound->Play();
	}




void StarsView :: fireEnemyBullet( Point location, unsigned int type )
	{
	Bullet b;
	Color32_s c( 250, 50, 50 ); /* Default bullet to red */
	b.int_color = 1;
	if( type == 2 ) /* Turn bullet yellow if ship is in yellow mode */
		{
		c.green = 250;
		b.int_color = 2;
		}
	b.color = c;
	b.location = location;
	b.velocity.x = -10;
	b.velocity.y = 0;
	b.type = type;
	bullets.push_back( b );
	}




void StarsView :: updateShipType( void )
	{
	/* Toggle between red and yellow modes */
	if( playerShipType == 1 )
		{
		playerImage = playerRedImage;
		}
	else
		{
		playerImage = playerYellowImage;
		}
	}




// ---------- Stars Window



class GameWindow : public Window
	{
	public:
		GameWindow( const Rect& r ); /* Constructor */
		void HandleMessage( Message* pcMessage );
		bool OkToQuit();
	private:
		StarsView *view;
		MoveStarsThread *thread;
	};




GameWindow :: GameWindow( const Rect& r )
	: Window( r, "FirestormWindow", "Firestorm", WND_NOT_RESIZABLE |
	          WND_NO_ZOOM_BUT | WND_NO_DEPTH_BUT | WND_SINGLEBUFFER,
	          CURRENT_DESKTOP )
	{
	window = this;
	view = new StarsView( GetBounds() );
	AddChild( view );
	SetFocusChild( view );
	thread = new MoveStarsThread();
	thread->Start();
	}




void GameWindow :: HandleMessage( Message *pcMessage )
	{
	switch( pcMessage->GetCode() )
		{
		case MSG_FIRE_ENEMY_BULLET:
			{
printf( "Got \"Fire Enemy Bullet\" message\n" );
			Point location;
			int32 type;
			pcMessage->FindPoint( "location", &location );
			pcMessage->FindInt32( "type", &type );
			view->fireEnemyBullet( location, (unsigned int)type );
			}
			break;
		case MSG_MOVE_STARS:
			view->DrawStars();
			break;
		default:
			Window :: HandleMessage( pcMessage );
		}
	}




bool GameWindow :: OkToQuit( void )
	{
	thread->Terminate();
	Application::GetInstance()->PostMessage( M_QUIT );
	return( true );
	}




// ---------- Firestorm Application




class MainApplication : public Application
	{
	public:
		MainApplication( void );
	private:
		GameWindow *win;
	};




MainApplication :: MainApplication( void )
		: Application( "application/x-vind.STUN-Firestorm" )
	{
	win = new GameWindow( Rect( 20, 40, 700, 340 ) );
	win->Show();
	win->MakeFocus();
	}




// ---------- main()




int main( void )
	{
	MainApplication *thisApp;
	thisApp = new MainApplication();
	thisApp->Run();
	return( 0 );
	}






